import yaml
import datetime
import os.path

from pydicti import odicti

from .utils import StatusTypes

from .calculators.cycletime import CycleTimeCalculator
from .calculators.cfd import CFDCalculator
from .calculators.scatterplot import ScatterplotCalculator
from .calculators.histogram import HistogramCalculator
from .calculators.percentiles import PercentilesCalculator
from .calculators.throughput import ThroughputCalculator
from .calculators.burnup import BurnupCalculator
from .calculators.wip import WIPChartCalculator
from .calculators.netflow import NetFlowChartCalculator
from .calculators.ageingwip import AgeingWIPChartCalculator
from .calculators.forecast import BurnupForecastCalculator
from .calculators.impediments import ImpedimentsCalculator
from .calculators.debt import DebtCalculator
from .calculators.defects import DefectsCalculator
from .calculators.waste import WasteCalculator

CALCULATORS = (
    CycleTimeCalculator,  # should come first -- others depend on results from this one
    CFDCalculator,        # needs to come before burn-up charts, wip charts, and net flow charts
    ScatterplotCalculator,
    HistogramCalculator,
    PercentilesCalculator,
    ThroughputCalculator,
    BurnupCalculator,
    WIPChartCalculator,
    NetFlowChartCalculator,
    AgeingWIPChartCalculator,
    BurnupForecastCalculator,
    ImpedimentsCalculator,
    DebtCalculator,
    DefectsCalculator,
    WasteCalculator,
)

class ConfigError(Exception):
    pass

# From http://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=odicti):
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping
    )

    return yaml.load(stream, OrderedLoader)

def force_list(val):
    return list(val) if isinstance(val, (list, tuple,)) else [val]

def expand_key(key):
    return str(key).replace('_', ' ').lower()

def config_to_options(data):
    try:
        config = ordered_load(data, yaml.SafeLoader)
    except Exception:
        raise ConfigError("Unable to parse YAML configuration file.") from None

    options = {
        'connection': {
            'domain': None,
            'username': None,
            'password': None,
            'jira_client_options': {}
        },
        'settings': {
            'queries': [],
            'query_attribute': None,
            'attributes': {},
            'known_values': {},
            'cycle': [],
            'max_results': None,
            'verbose': False,
        
            'quantiles': [0.5, 0.85, 0.95],

            'backlog_column': None,
            'committed_column': None,
            'final_column': None,
            'done_column': None,

            'cycle_time_data': None,
            'percentiles_data': None,

            'scatterplot_window': None,
            'scatterplot_data': None,
            'scatterplot_chart': None,
            'scatterplot_chart_title': None,
            
            'histogram_window': None,
            'histogram_data': None,
            'histogram_chart': None,
            'histogram_chart_title': None,

            'cfd_window': None,
            'cfd_data': None,
            'cfd_chart': None,
            'cfd_chart_title': None,
            
            'throughput_frequency': '1W-MON',
            'throughput_window': None,
            'throughput_data': None,
            'throughput_chart': None,
            'throughput_chart_title': None,
            
            'burnup_window': None,
            'burnup_chart': None,
            'burnup_chart_title': None,

            'burnup_forecast_window': None,
            'burnup_forecast_chart': None,
            'burnup_forecast_chart_title': None,
            'burnup_forecast_chart_target': None,
            'burnup_forecast_chart_deadline': None,
            'burnup_forecast_chart_deadline_confidence': None,
            'burnup_forecast_chart_trials': 100,
            'burnup_forecast_chart_throughput_window': 60,
            'burnup_forecast_chart_throughput_window_end': None,

            'wip_frequency': '1W-MON',
            'wip_window': None,
            'wip_chart': None,
            'wip_chart_title': None,

            'ageing_wip_chart': None,
            'ageing_wip_chart_title': None,

            'net_flow_frequency': '1W-MON',
            'net_flow_window': None,
            'net_flow_chart': None,
            'net_flow_chart_title': None,

            'impediments_data': None,
            'impediments_window': None,
            'impediments_chart': None,
            'impediments_chart_title': None,
            'impediments_days_chart': None,
            'impediments_days_chart_title': None,
            
            'defects_query': None,
            'defects_window': None,
            'defects_priority_field': None,
            'defects_priority_values': None,
            'defects_type_field': None,
            'defects_type_values': None,
            'defects_environment_field': None,
            'defects_environment_values': None,

            'defects_by_priority_chart': None,
            'defects_by_priority_chart_title': None,
            'defects_by_type_chart': None,
            'defects_by_type_chart_title': None,
            'defects_by_environment_chart': None,
            'defects_by_environment_chart_title': None,
        
            'debt_query': None,
            'debt_window': None,
            'debt_priority_field': None,
            'debt_priority_values': None,
            'debt_chart': None,
            'debt_chart_title': None,
            'debt_age_chart': None,
            'debt_age_chart_title': None,
            'debt_age_chart_bins': [30, 60, 90],

            'waste_query': None,
            'waste_window': None,
            'waste_frequency': 'MS',
            'waste_chart': None,
            'waste_chart_title': None,
        }
    }

    # Parse and validate Connection

    if 'domain' in config['connection']:
        options['connection']['domain'] = config['connection']['domain']

    if 'username' in config['connection']:
        options['connection']['username'] = config['connection']['username']

    if 'password' in config['connection']:
        options['connection']['password'] = config['connection']['password']

    if 'jira_client_options' in config['connection']:
        options['connection']['jira_client_options'] = config['connection']['jira_client_options']

    # Parse and validate output options
    if 'output' in config:

        if 'quantiles' in config['output']:
            try:
                options['settings']['quantiles'] = list(map(float, config['output']['quantiles']))
            except ValueError:
                raise ConfigError("Could not convert value `%s` for key `quantiles` to a list of decimals" % (config['output']['quantiles'],)) from None

        # int values
        for key in [
            'scatterplot_window',
            'histogram_window',
            'wip_window',
            'net_flow_window',
            'throughput_window',
            'cfd_window',
            'burnup_window',
            'burnup_forecast_window',
            'burnup_forecast_chart_throughput_window',
            'burnup_forecast_chart_target',
            'burnup_forecast_chart_trials',
            'impediments_window',
            'defects_window',
            'debt_window',
            'waste_window',
        ]:
            if expand_key(key) in config['output']:
                value = config['output'][expand_key(key)]
                try:
                    options['settings'][key] = int(value)
                except ValueError:
                    raise ConfigError("Could not convert value `%s` for key `%s` to integer" % (value, expand_key(key),)) from None
        
        # float values
        for key in [
            'burnup_forecast_chart_deadline_confidence',
        ]:
            if expand_key(key) in config['output']:
                value = config['output'][expand_key(key)]
                try:
                    options['settings'][key] = float(value)
                except ValueError:
                    raise ConfigError("Could not convert value `%s` for key `%s` to decimal" % (value, expand_key(key),)) from None

        # date values
        for key in [
            'burnup_forecast_chart_throughput_window_end',
            'burnup_forecast_chart_deadline',
        ]:
            if expand_key(key) in config['output']:
                value = config['output'][expand_key(key)]
                if not isinstance(value, datetime.date):
                    raise ConfigError("Value `%s` for key `%s` is not a date" % (value, expand_key(key),))
                options['settings'][key] = value
        
        # file name values
        for key in [
            'cycle_time_data',
            'cfd_data',
            'scatterplot_data',
            'histogram_data',
            'throughput_data',
            'percentiles_data',
            'scatterplot_chart',
            'histogram_chart',
            'cfd_chart',
            'throughput_chart',
            'burnup_chart',
            'burnup_forecast_chart',
            'wip_chart',
            'ageing_wip_chart',
            'net_flow_chart',
            'impediments_data',
            'impediments_chart',
            'impediments_days_chart',
            'defects_by_priority_chart',
            'defects_by_type_chart',
            'defects_by_environment_chart',
            'debt_chart',
            'debt_age_chart',
            'waste_chart',
        ]:
            if expand_key(key) in config['output']:
                options['settings'][key] = os.path.basename(config['output'][expand_key(key)])

        # list values
        for key in [
            'defects_priority_values',
            'defects_type_values',
            'defects_environment_values',
            'debt_priority_values',
            'debt_age_chart_bins',
        ]:
            if expand_key(key) in config['output']:
                options['settings'][key] = force_list(config['output'][expand_key(key)])

        # string values that copy straight over
        for key in [
            'backlog_column',
            'committed_column',
            'final_column',
            'done_column',
            'throughput_frequency',
            'scatterplot_chart_title',
            'histogram_chart_title',
            'cfd_chart_title',
            'throughput_chart_title',
            'burnup_chart_title',
            'burnup_forecast_chart_title',
            'wip_chart_title',
            'wip_frequency',
            'ageing_wip_chart_title',
            'net_flow_chart_title',
            'net_flow_frequency',
            'impediments_chart_title',
            'impediments_days_chart_title',
            'defects_query',
            'defects_by_priority_chart_title',
            'defects_priority_field',
            'defects_by_type_chart_title',
            'defects_type_field',
            'defects_by_environment_chart_title',
            'defects_environment_field',
            'debt_query',
            'debt_priority_field',
            'debt_chart_title',
            'debt_age_chart_title',
            'waste_query',
            'waste_frequency',
            'waste_chart_title',
        ]:
            if expand_key(key) in config['output']:
                options['settings'][key] = config['output'][expand_key(key)]

    # Parse Queries and/or a single Query

    if 'queries' in config:
        options['settings']['query_attribute'] = config['queries'].get('attribute', None)
        for query in config['queries']['criteria']:
            options['settings']['queries'].append({
                'value': query.get('value', None),
                'jql': query.get('jql', None),
            })

    if 'query' in config:
        options['settings']['queries'].append({
            'value': None,
            'jql': config['query'],
        })

    if len(options['settings']['queries']) == 0:
        raise ConfigError("No `Query` value or `Queries` section found")

    # Parse Workflow. Assume first status is backlog and last status is complete.

    if 'workflow' not in config:
        raise ConfigError("`Workflow` section not found")

    if len(config['workflow'].keys()) < 3:
        raise ConfigError("`Workflow` section must contain at least three statuses")

    for name, statuses in config['workflow'].items():
        statuses = force_list(statuses)

        options['settings']['cycle'].append({
            "name": name,
            "type": StatusTypes.accepted,
            "statuses": statuses
        })

    options['settings']['cycle'][0]['type'] = StatusTypes.backlog
    options['settings']['cycle'][-1]['type'] = StatusTypes.complete

    if options['settings']['backlog_column'] is None:
        options['settings']['backlog_column'] = options['settings']['cycle'][0]['name']
    if options['settings']['committed_column'] is None:
        options['settings']['committed_column'] = options['settings']['cycle'][1]['name']
    if options['settings']['final_column'] is None:
        options['settings']['final_column'] = options['settings']['cycle'][-2]['name']
    if options['settings']['done_column'] is None:
        options['settings']['done_column'] = options['settings']['cycle'][-1]['name']

    # Parse attributes (fields)

    if 'attributes' in config:
        options['settings']['attributes'] = dict(config['attributes'])

    if 'known values' in config:
        for name, values in config['known values'].items():
            options['settings']['known_values'][name] = force_list(values)

    return options



import pytest
from pandas import Timestamp

from ..conftest import (
    FauxJIRA as JIRA,
    FauxIssue as Issue,
    FauxFieldValue as Value,
    FauxChange as Change
)

from ..utils import extend_dict

from ..querymanager import QueryManager
from .waste import WasteCalculator

@pytest.fixture
def fields(minimal_fields):
    return minimal_fields + []

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
        'waste_query': 'issueType = Story AND resolution IN (Withdrawn, Invalid)',
        'waste_window': 3,
        'waste_frequency': '2W-WED',
        'waste_chart': 'waste.png',
        'waste_chart_title': 'Waste',
    })

@pytest.fixture
def jira(fields):
    return JIRA(fields=fields, issues=[
        Issue("A-1",
            summary="Withdrawn from QA",
            issuetype=Value("Story", "story"),
            status=Value("Done", "done"),
            created="2018-01-03 01:01:01",
            resolution=Value("Withdrawn", "withdrawn"),
            resolutiondate="2018-01-06 02:02:02",
            changes=[
                Change("2018-01-03 02:02:02", [("status", "Backlog", "Next",)]),
                Change("2018-01-04 02:02:02", [("status", "Next", "Build",)]),
                Change("2018-01-05 02:02:02", [("status", "Build", "QA",)]),
                Change("2018-01-06 02:02:02", [("status", "QA", "Done",)]),
            ],
        ),
        Issue("A-2",
            summary="Withdrawn from Next",
            issuetype=Value("Story", "story"),
            status=Value("Done", "done"),
            created="2018-01-03 01:01:01",
            resolution=Value("Withdrawn", "withdrawn"),
            resolutiondate="2018-01-07 02:02:02",
            changes=[
                Change("2018-01-03 02:02:02", [("status", "Backlog", "Next",)]),
                Change("2018-01-07 02:02:02", [("status", "Next", "Done",)]),
            ],
        ),
        Issue("A-3",
            summary="Withdrawn from Done",
            issuetype=Value("Story", "story"),
            status=Value("Done", "done"),
            created="2018-01-03 01:01:01",
            resolution=Value("Withdrawn", "withdrawn"),
            resolutiondate="2018-01-08 02:02:02",
            changes=[
                Change("2018-01-03 02:02:02", [("status", "Backlog", "Next",)]),
                Change("2018-01-04 02:02:02", [("status", "Next", "Build",)]),
                Change("2018-01-05 02:02:02", [("status", "Build", "QA",)]),
                Change("2018-01-06 02:02:02", [("status", "QA", "Done",)]),
                Change("2018-01-08 02:02:02", [("status", "Done", "Done",)]),
            ],
        ),
        Issue("A-4",
            summary="Withdrawn from Backlog",
            issuetype=Value("Story", "story"),
            status=Value("Done", "done"),
            created="2018-01-03 01:01:01",
            resolution=Value("Withdrawn", "withdrawn"),
            resolutiondate="2018-01-09 02:02:02",
            changes=[
                Change("2018-01-09 02:02:02", [("status", "Backlog", "Done",)]),
            ],
        ),
        Issue("A-5",
            summary="Unresolved",
            issuetype=Value("Story", "story"),
            status=Value("Done", "done"),
            created="2018-01-03 01:01:01",
            resolution=None,
            resolutiondate=None,
            changes=[
                Change("2018-01-03 02:02:02", [("status", "Backlog", "Next",)]),
                Change("2018-01-04 02:02:02", [("status", "Next", "Build",)]),
                Change("2018-01-05 02:02:02", [("status", "Build", "QA",)]),
                Change("2018-01-06 02:02:02", [("status", "QA", "Done",)]),
            ],
        ),
        Issue("A-6",
            summary="Unmapped state",
            issuetype=Value("Story", "story"),
            status=Value("Done", "done"),
            created="2018-01-03 01:01:01",
            resolution=Value("Withdrawn", "withdrawn"),
            resolutiondate="2018-01-06 02:02:02",
            changes=[
                Change("2018-01-03 02:02:02", [("status", "Backlog", "Next",)]),
                Change("2018-01-04 02:02:02", [("status", "Next", "Build",)]),
                Change("2018-01-05 02:02:02", [("status", "Build", "foobar",)]),
                Change("2018-01-06 02:02:02", [("status", "foobar", "Done",)]),
            ],
        ),
        Issue("A-7",
            summary="No changes",
            issuetype=Value("Story", "story"),
            status=Value("Done", "done"),
            created="2018-01-06 02:02:02",
            resolution=Value("Withdrawn", "withdrawn"),
            resolutiondate="2018-01-06 02:02:02",
            changes=[],
        ),
    ])

def test_no_query(jira, settings):
    query_manager = QueryManager(jira, settings)
    results = {}
    settings = extend_dict(settings, {
        'waste_query': None
    })
    calculator = WasteCalculator(query_manager, settings, results)

    data = calculator.run()
    assert data is None

def test_columns(jira, settings):
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = WasteCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.columns) == ['key', 'last_status', 'resolution', 'withdrawn_date']

def test_empty(fields, settings):
    jira = JIRA(fields=fields, issues=[])
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = WasteCalculator(query_manager, settings, results)

    data = calculator.run()

    assert len(data.index) == 0

def test_query(jira, settings):
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = WasteCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'A-1', 'last_status': 'Test',      'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-06 02:02:02')},
        {'key': 'A-2', 'last_status': 'Committed', 'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-07 02:02:02')},
        {'key': 'A-6', 'last_status': 'foobar',    'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-06 02:02:02')},
        {'key': 'A-7', 'last_status': None,        'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-06 02:02:02')},
    ]

def test_different_backlog_column(jira, settings):
    settings = extend_dict(settings, {
        'backlog_column': 'Committed'
    })

    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = WasteCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'A-1', 'last_status': 'Test',      'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-06 02:02:02')},
        {'key': 'A-4', 'last_status': 'Backlog',   'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-09 02:02:02')},
        {'key': 'A-6', 'last_status': 'foobar',    'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-06 02:02:02')},
        {'key': 'A-7', 'last_status': None,        'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-06 02:02:02')},
    ]

def test_different_done_column(jira, settings):
    settings = extend_dict(settings, {
        'done_column': 'Test'
    })

    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = WasteCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'A-2', 'last_status': 'Committed', 'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-07 02:02:02')},
        {'key': 'A-3', 'last_status': 'Done',      'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-08 02:02:02')},
        {'key': 'A-6', 'last_status': 'foobar',    'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-06 02:02:02')},
        {'key': 'A-7', 'last_status': None,        'resolution': 'Withdrawn', 'withdrawn_date': Timestamp('2018-01-06 02:02:02')},
    ]



import logging

logger = logging.getLogger(__name__)

class Calculator(object):
    """Base class for calculators.
    """

    def __init__(self, query_manager, settings, results):
        """Initialise with a `QueryManager`, a dict of `settings`,
        and a reference to the dict of `results`, which will be
        used to store intermeidary results.
        """

        self.query_manager = query_manager
        self.settings = settings
        self._results = results

    def get_result(self, calculator=None, default=None):
        """Get the results calculated by a previous calculator
        of type `calculator` (a class). Defaults to `self.__class__`
        """

        return self._results.get(calculator or self.__class__, default)

    # Lifecycle methods -- implement as appropriate

    def run(self):
        """Run the calculator and return its results.
        These will be automatically saved
        """

    def write(self):
        """Write any output files to the filesystem in the given
        target directory.
        """

def run_calculators(calculators, query_manager, settings):
    """Run all calculators passed in, in the order listed.
    Returns the aggregated results.
    """

    results = {}
    calculators = [C(query_manager, settings, results) for C in calculators]
    failed = set()

    # Run all calculators first
    for c in calculators:
        logger.info("%s running...", c.__class__.__name__)
        try:
            results[c.__class__] = c.run()
        except Exception as e:
            failed.add(c)
            logger.exception("%s failed with a fatal error. Subsequent calculators may also fail as a result.", c.__class__.__name__)
        else:
            logger.info("%s completed\n", c.__class__.__name__)

    # Write all files as a second pass
    for c in calculators:
        logger.info("Writing file for %s...", c.__class__.__name__)
        try:
            c.write()
        except Exception as e:
            logger.exception("%s failed with a fatal error. Subsequent calculators may also fail as a result.", c.__class__.__name__)
        else:
            logger.info("%s completed\n", c.__class__.__name__)

    return results



import logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from ..calculator import Calculator
from ..utils import get_extension, set_chart_style

from .cycletime import CycleTimeCalculator

logger = logging.getLogger(__name__)

class HistogramCalculator(Calculator):
    """Build histogram data for the cycle times in `cycle_data`. Returns
    a dictionary with keys `bin_values` and `bin_edges` of numpy arrays
    """

    def run(self):
        cycle_data = self.get_result(CycleTimeCalculator)
        
        values, edges = np.histogram(cycle_data['cycle_time'].astype('timedelta64[D]').dropna(), bins=10)

        index = []
        for i, _ in enumerate(edges):
            if i == 0:
                continue
            index.append("%.01f to %.01f" % (edges[i - 1], edges[i],))

        return pd.Series(values, name="Items", index=index)
    
    def write(self):
        data = self.get_result()

        if self.settings['histogram_data']:
            self.write_file(data, self.settings['histogram_data'])
        else:
            logger.debug("No output file specified for histogram data")
        
        if self.settings['histogram_chart']:
            self.write_chart(data, self.settings['histogram_chart'])
        else:
            logger.debug("No output file specified for histogram chart")

    def write_file(self, data, output_file):
        output_extension = get_extension(output_file)

        file_data = self.get_result()

        logger.info("Writing histogram data to %s", output_file)
        if output_extension == '.json':
            file_data.to_json(output_file, date_format='iso')
        elif output_extension == '.xlsx':
            file_data.to_frame(name='histogram').to_excel(output_file, 'Histogram', header=True)
        else:
            file_data.to_csv(output_file, header=True)
    
    def write_chart(self, data, output_file):
        cycle_data = self.get_result(CycleTimeCalculator)
        chart_data = cycle_data[['cycle_time', 'completed_timestamp']].dropna(subset=['cycle_time'])

        # The `window` calculation and the chart output will fail if we don't
        # have at least two valid data points.
        ct_days = chart_data['cycle_time'].dt.days
        if len(ct_days.index) < 2:
            logger.warning("Need at least 2 completed items to draw histogram")
            return

        # Slice off items before the window
        window = self.settings['histogram_window']
        if window:
            start = chart_data['completed_timestamp'].max().normalize() - pd.Timedelta(window, 'D')
            chart_data = chart_data[chart_data.completed_timestamp >= start]
        
            # Re-check that we have enough data
            ct_days = chart_data['cycle_time'].dt.days
            if len(ct_days.index) < 2:
                logger.warning("Need at least 2 completed items to draw histogram")
                return
        
        quantiles = self.settings['quantiles']
        logger.debug("Showing histogram at quantiles %s", ', '.join(['%.2f' % (q * 100.0) for q in quantiles]))

        fig, ax = plt.subplots()

        sns.distplot(ct_days, bins=30, ax=ax, axlabel="Cycle time (days)")
        
        if self.settings['histogram_chart_title']:
            ax.set_title(self.settings['histogram_chart_title'])

        _, right = ax.get_xlim()
        ax.set_xlim(0, right)

        # Add quantiles
        bottom, top = ax.get_ylim()
        for quantile, value in ct_days.quantile(quantiles).iteritems():
            ax.vlines(value, bottom, top - 0.001, linestyles='--', linewidths=1)
            ax.annotate("%.0f%% (%.0f days)" % ((quantile * 100), value,),
                xy=(value, top),
                xytext=(value - 0.1, top - 0.001),
                rotation="vertical",
                fontsize="x-small",
                ha="right"
            )

        set_chart_style()
        
        # Write file
        logger.info("Writing histogram chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from ..calculator import Calculator
from ..utils import set_chart_style

from .cycletime import CycleTimeCalculator

logger = logging.getLogger(__name__)

class AgeingWIPChartCalculator(Calculator):
    """Draw an ageing WIP chart
    """

    def run(self, today=None):

        # short circuit relatively expensive calculation if it won't be used
        if not self.settings['ageing_wip_chart']:
            return None

        cycle_data = self.get_result(CycleTimeCalculator)
        cycle_names = [s['name'] for s in self.settings['cycle']]

        start_column = self.settings['committed_column']
        end_column = self.settings['final_column']
        done_column = self.settings['done_column']
        
        if start_column not in cycle_names:
            logger.error("Committed column %s does not exist", start_column)
            return None
        if end_column not in cycle_names:
            logger.error("Final column %s does not exist", end_column)
            return None
        if done_column not in cycle_names:
            logger.error("Done column %s does not exist", done_column)
            return None

        today = pd.Timestamp.now().date() if today is None else today  # to allow testing

        # remove items that are done
        ageing_wip_data = cycle_data[pd.isnull(cycle_data[done_column])].copy()

        # calculate current status and age for each item
        def extract_status(row):
            last_valid = row.last_valid_index()
            if last_valid is None:
                return np.NaN
            return last_valid

        def extract_age(row):
            if start_column not in row:
                return np.NaN
            started = row[start_column]
            if pd.isnull(started):
                return np.NaN
            return (today - started.date()).days

        ageing_wip_data['status'] = ageing_wip_data.apply(extract_status, axis=1)
        ageing_wip_data['age'] = ageing_wip_data.apply(extract_age, axis=1)

        # remove blank rows
        ageing_wip_data.dropna(how='any', inplace=True, subset=['status', 'age'])

        # reorder columns so we get key, summary, status, age, and then all the cycle stages
        ageing_wip_data = pd.concat((
            ageing_wip_data[['key', 'summary', 'status', 'age']],
            ageing_wip_data.loc[:, start_column:end_column]
        ), axis=1)

        return ageing_wip_data
    
    def write(self):
        output_file = self.settings['ageing_wip_chart']
        if not output_file:
            logger.debug("No output file specified for ageing WIP chart")
            return

        chart_data = self.get_result()

        if len(chart_data.index) == 0:
            logger.warning("Unable to draw ageing WIP chart with zero completed items")
            return

        fig, ax = plt.subplots()
        
        if self.settings['ageing_wip_chart_title']:
            ax.set_title(self.settings['ageing_wip_chart_title'])

        sns.swarmplot(x='status', y='age', order=chart_data.columns[4:], data=chart_data, ax=ax)

        ax.set_xlabel("Status")
        ax.set_ylabel("Age (days)")

        ax.set_xticklabels(ax.xaxis.get_majorticklabels(), rotation=90)

        _, top = ax.get_ylim()
        ax.set_ylim(0, top)

        set_chart_style()

        # Write file
        logger.info("Writing ageing WIP chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import logging
import matplotlib.pyplot as plt

from ..calculator import Calculator
from ..utils import set_chart_style

from .cfd import CFDCalculator

logger = logging.getLogger(__name__)

class NetFlowChartCalculator(Calculator):
    """Draw a net flow chart
    """

    def run(self):
        cfd_data = self.get_result(CFDCalculator)
        cycle_names = [s['name'] for s in self.settings['cycle']]

        start_column = self.settings['committed_column']
        done_column = self.settings['done_column']

        if start_column not in cycle_names:
            logger.error("Committed column %s does not exist", start_column)
            return None
        if done_column not in cycle_names:
            logger.error("Done column %s does not exist", done_column)
            return None

        frequency = self.settings['net_flow_frequency']
        logger.debug("Calculating net flow at frequency %s", frequency)
        
        net_flow_data = cfd_data[[start_column, done_column]].resample(frequency, label='left').max()
        net_flow_data['arrivals'] = net_flow_data[start_column].diff().fillna(net_flow_data[start_column])
        net_flow_data['departures'] = net_flow_data[done_column].diff().fillna(net_flow_data[done_column])
        net_flow_data['net_flow'] = net_flow_data['arrivals'] - net_flow_data['departures']
        net_flow_data['positive'] = net_flow_data['net_flow'] >= 0

        return net_flow_data
    
    def write(self):
        output_file = self.settings['net_flow_chart']
        if not output_file:
            logger.debug("No output file specified for net flow chart")
            return

        chart_data = self.get_result()

        if len(chart_data.index) == 0:
            logger.warning("Cannot draw net flow chart with zero items")
            return

        fig, ax = plt.subplots()
        
        if self.settings['net_flow_chart_title']:
            ax.set_title(self.settings['net_flow_chart_title'])

        ax.set_xlabel("Period starting")
        ax.set_ylabel("Net flow (departures - arrivals)")

        net_flow_data = chart_data[['net_flow', 'positive']]

        window = self.settings['net_flow_window']
        if window:
            net_flow_data = net_flow_data[-window:]

        net_flow_data['net_flow'].plot.bar(ax=ax, color=net_flow_data['positive'].map({True: 'r', False: 'b'}),)

        labels = [d.strftime("%d/%m/%Y") for d in net_flow_data.index]
        ax.set_xticklabels(labels, rotation=70, size='small')

        set_chart_style()

        # Write file
        logger.info("Writing ageing WIP chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import logging
import datetime
import dateutil.parser

import pandas as pd
import matplotlib.pyplot as plt

from ..calculator import Calculator
from ..utils import breakdown_by_month, set_chart_style, to_bin

logger = logging.getLogger(__name__)

class DebtCalculator(Calculator):
    """Calculate technical debt over time.

    Queries JIRA with JQL set in `debt_query` and draws a stacked bar
    chart in the file `debt_chart` with title `debt_chart_title`. The bars are
    the last 6 months (or another window set in `debt_window`), grouped by
    priority. The field giving the priority is set with `debt_chart_priority`.
    To force the list of valid values and their order, provide a list of strings
    in `debt_priority_values`.

    Also draw a stacked bar chart in the file `debt_age_chart`, with title
    `debt_age_chart_title`, grouping by item age.
    """

    def run(self, now=None):

        query = self.settings['debt_query']

        # Allows unit testing to use a fixed date
        if now is None:
            now = datetime.datetime.utcnow()

        # This calculation is expensive. Only run it if we have a query.
        if not query:
            logger.debug("Not calculating debt chart data as no query specified")
            return None
        
        # Resolve field name to field id for later lookup
        priority_field = self.settings['debt_priority_field']
        priority_field_id = priority_field_id = self.query_manager.field_name_to_id(priority_field) if priority_field else None

        # Build data frame
        columns = ['key', 'priority', 'created', 'resolved', 'age']
        series = {
            'key': {'data': [], 'dtype': 'str'},
            'priority': {'data': [], 'dtype': 'str'},
            'created': {'data': [], 'dtype': 'datetime64[ns]'},
            'resolved': {'data': [], 'dtype': 'datetime64[ns]'},
            'age': {'data': [], 'dtype': 'timedelta64[ns]'},
        }

        for issue in self.query_manager.find_issues(query, expand=None):
            created_date = dateutil.parser.parse(issue.fields.created)
            resolved_date = dateutil.parser.parse(issue.fields.resolutiondate) if issue.fields.resolutiondate else None

            series['key']['data'].append(issue.key)
            series['priority']['data'].append(self.query_manager.resolve_field_value(issue, priority_field_id) if priority_field else None)
            series['created']['data'].append(created_date)
            series['resolved']['data'].append(resolved_date)
            series['age']['data'].append((resolved_date.replace(tzinfo=None) if resolved_date is not None else now) - created_date.replace(tzinfo=None))

        data = {}
        for k, v in series.items():
            data[k] = pd.Series(v['data'], dtype=v['dtype'])

        return pd.DataFrame(data, columns=columns)

    def write(self):
        chart_data = self.get_result()
        if chart_data is None:
            return

        if len(chart_data.index) == 0:
            logger.warning("Cannot draw debt chart with zero items")
            return
        
        if self.settings['debt_chart']:
            self.write_debt_chart(chart_data, self.settings['debt_chart'])
        
        if self.settings['debt_age_chart']:
            self.write_debt_age_chart(chart_data, self.settings['debt_age_chart'])
    
    def write_debt_chart(self, chart_data, output_file):
        window = self.settings['debt_window']
        priority_values = self.settings['debt_priority_values']

        breakdown = breakdown_by_month(chart_data, 'created', 'resolved', 'key', 'priority', priority_values)
        
        if window:
            breakdown = breakdown[-window:]

        fig, ax = plt.subplots()
        
        breakdown.plot.bar(ax=ax, stacked=True)
        
        if self.settings['debt_chart_title']:
            ax.set_title(self.settings['debt_chart_title'])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("Month", labelpad=20)
        ax.set_ylabel("Number of items", labelpad=10)

        labels = [d.strftime("%b %y") for d in breakdown.index]
        ax.set_xticklabels(labels, rotation=90, size='small')

        set_chart_style()

        # Write file
        logger.info("Writing debt chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)
    
    def write_debt_age_chart(self, chart_data, output_file):
        priority_values = self.settings['debt_priority_values']
        bins = self.settings['debt_age_chart_bins']
        
        def generate_bin_label(v):
            low, high = to_bin(v, bins)
            return "> %d days" % (low,) if high is None else "%d-%d days" % (low, high,)

        def day_grouper(value):
            if isinstance(value, pd.Timedelta):
                return generate_bin_label(value.days)

        bin_labels = list(map(generate_bin_label, bins + [bins[-1] + 1]))
        breakdown = chart_data.pivot_table(
            index='age',
            columns='priority',
            values='key',
            aggfunc='count'
        ).groupby(day_grouper).sum().reindex(bin_labels).T
        
        if priority_values:
            breakdown = breakdown.reindex(priority_values)

        fig, ax = plt.subplots()
        
        breakdown.plot.barh(ax=ax, stacked=True)
        
        if self.settings['debt_age_chart_title']:
            ax.set_title(self.settings['debt_age_chart_title'])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("Number of items", labelpad=20)
        ax.set_ylabel("Priority", labelpad=10)

        set_chart_style()

        # Write file
        logger.info("Writing debt age chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import datetime
import pytest
from pandas import Timedelta, Timestamp, NaT

from ..conftest import (
    FauxJIRA as JIRA,
    FauxIssue as Issue,
    FauxFieldValue as Value
)

from ..utils import extend_dict

from ..querymanager import QueryManager
from .defects import DefectsCalculator

@pytest.fixture
def fields(minimal_fields):
    return minimal_fields + [
        {'id': 'priority',  'name': 'Priority'},
        {'id': 'customfield_001',  'name': 'Environment'},
        {'id': 'customfield_002',  'name': 'Defect type'},
    ]

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
        'defects_query': 'issueType = Defect',
        'defects_window': 3,
        'defects_priority_field': 'Priority',
        'defects_priority_values': ['Low', 'Medium', 'High'],
        'defects_type_field': 'Defect type',
        'defects_type_values': ['Config', 'Data', 'Code'],
        'defects_environment_field': 'Environment',
        'defects_environment_values': ['SIT', 'UAT', 'PROD'],

        'defects_by_priority_chart': 'defects-by-priority.png',
        'defects_by_priority_chart_title': 'Defects by priority',
        'defects_by_type_chart': 'defects-by-type.png',
        'defects_by_type_chart_title': 'Defects by type',
        'defects_by_environment_chart': 'defects-by-environment.png',
        'defects_by_environment_chart_title': 'Defects by environment',
    })

@pytest.fixture
def jira(fields):
    return JIRA(fields=fields, issues=[
        Issue("D-1",
            summary="Debt 1",
            issuetype=Value("Bug", "Bug"),
            status=Value("Closed", "closed"),
            created="2018-01-01 01:01:01",
            resolution="Done",
            resolutiondate="2018-03-20 02:02:02",
            priority=Value("High", "High"),
            customfield_001=Value(None, "PROD"),
            customfield_002=Value(None, "Config"),
            changes=[],
        ),
        Issue("D-2",
            summary="Debt 2",
            issuetype=Value("Bug", "Bug"),
            status=Value("Closed", "closed"),
            created="2018-01-02 01:01:01",
            resolution="Done",
            resolutiondate="2018-01-20 02:02:02",
            priority=Value("Medium", "Medium"),
            customfield_001=Value(None, "SIT"),
            customfield_002=Value(None, "Config"),
            changes=[],
        ),
        Issue("D-3",
            summary="Debt 3",
            issuetype=Value("Bug", "Bug"),
            status=Value("Closed", "closed"),
            created="2018-02-03 01:01:01",
            resolution="Done",
            resolutiondate="2018-03-20 02:02:02",
            priority=Value("High", "High"),
            customfield_001=Value(None, "UAT"),
            customfield_002=Value(None, "Config"),
            changes=[],
        ),
        Issue("D-4",
            summary="Debt 4",
            issuetype=Value("Bug", "Bug"),
            status=Value("Closed", "closed"),
            created="2018-01-04 01:01:01",
            resolution=None,
            resolutiondate=None,
            priority=Value("Medium", "Medium"),
            customfield_001=Value(None, "PROD"),
            customfield_002=Value(None, "Data"),
            changes=[],
        ),
        Issue("D-5",
            summary="Debt 5",
            issuetype=Value("Bug", "Bug"),
            status=Value("Closed", "closed"),
            created="2018-02-05 01:01:01",
            resolution="Done",
            resolutiondate="2018-02-20 02:02:02",
            priority=Value("High", "High"),
            customfield_001=Value(None, "SIT"),
            customfield_002=Value(None, "Data"),
            changes=[],
        ),
        Issue("D-6",
            summary="Debt 6",
            issuetype=Value("Bug", "Bug"),
            status=Value("Closed", "closed"),
            created="2018-03-06 01:01:01",
            resolution=None,
            resolutiondate=None,
            priority=Value("Medium", "Medium"),
            customfield_001=Value(None, "UAT"),
            customfield_002=Value(None, "Data"),
            changes=[],
        ),
    ])

def test_no_query(jira, settings):
    query_manager = QueryManager(jira, settings)
    results = {}
    settings = extend_dict(settings, {
        'defects_query': None
    })
    calculator = DefectsCalculator(query_manager, settings, results)

    data = calculator.run()
    assert data is None

def test_columns(jira, settings):
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = DefectsCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.columns) == ['key', 'priority', 'type', 'environment', 'created', 'resolved']

def test_empty(fields, settings):
    jira = JIRA(fields=fields, issues=[])
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = DefectsCalculator(query_manager, settings, results)

    data = calculator.run()

    assert len(data.index) == 0


def test_breakdown(jira, settings):
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = DefectsCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'D-1', 'created': Timestamp('2018-01-01 01:01:01'), 'resolved': Timestamp('2018-03-20 02:02:02'), 'priority': 'High',   'environment': 'PROD', 'type': 'Config'},
        {'key': 'D-2', 'created': Timestamp('2018-01-02 01:01:01'), 'resolved': Timestamp('2018-01-20 02:02:02'), 'priority': 'Medium', 'environment': 'SIT',  'type': 'Config'},
        {'key': 'D-3', 'created': Timestamp('2018-02-03 01:01:01'), 'resolved': Timestamp('2018-03-20 02:02:02'), 'priority': 'High',   'environment': 'UAT',  'type': 'Config'},
        {'key': 'D-4', 'created': Timestamp('2018-01-04 01:01:01'), 'resolved': NaT,                              'priority': 'Medium', 'environment': 'PROD', 'type': 'Data'},
        {'key': 'D-5', 'created': Timestamp('2018-02-05 01:01:01'), 'resolved': Timestamp('2018-02-20 02:02:02'), 'priority': 'High',   'environment': 'SIT',  'type': 'Data'},
        {'key': 'D-6', 'created': Timestamp('2018-03-06 01:01:01'), 'resolved': NaT,                              'priority': 'Medium', 'environment': 'UAT',  'type': 'Data'},
    ]


def test_no_priority_field(jira, settings):
    settings = extend_dict(settings, {
        'defects_priority_field': None
    })

    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = DefectsCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'D-1', 'created': Timestamp('2018-01-01 01:01:01'), 'resolved': Timestamp('2018-03-20 02:02:02'), 'priority': None, 'environment': 'PROD', 'type': 'Config'},
        {'key': 'D-2', 'created': Timestamp('2018-01-02 01:01:01'), 'resolved': Timestamp('2018-01-20 02:02:02'), 'priority': None, 'environment': 'SIT',  'type': 'Config'},
        {'key': 'D-3', 'created': Timestamp('2018-02-03 01:01:01'), 'resolved': Timestamp('2018-03-20 02:02:02'), 'priority': None, 'environment': 'UAT',  'type': 'Config'},
        {'key': 'D-4', 'created': Timestamp('2018-01-04 01:01:01'), 'resolved': NaT,                              'priority': None, 'environment': 'PROD', 'type': 'Data'},
        {'key': 'D-5', 'created': Timestamp('2018-02-05 01:01:01'), 'resolved': Timestamp('2018-02-20 02:02:02'), 'priority': None, 'environment': 'SIT',  'type': 'Data'},
        {'key': 'D-6', 'created': Timestamp('2018-03-06 01:01:01'), 'resolved': NaT,                              'priority': None, 'environment': 'UAT',  'type': 'Data'},
    ]

def test_no_type_field(jira, settings):
    settings = extend_dict(settings, {
        'defects_type_field': None
    })

    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = DefectsCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'D-1', 'created': Timestamp('2018-01-01 01:01:01'), 'resolved': Timestamp('2018-03-20 02:02:02'), 'priority': 'High',   'environment': 'PROD', 'type': None},
        {'key': 'D-2', 'created': Timestamp('2018-01-02 01:01:01'), 'resolved': Timestamp('2018-01-20 02:02:02'), 'priority': 'Medium', 'environment': 'SIT',  'type': None},
        {'key': 'D-3', 'created': Timestamp('2018-02-03 01:01:01'), 'resolved': Timestamp('2018-03-20 02:02:02'), 'priority': 'High',   'environment': 'UAT',  'type': None},
        {'key': 'D-4', 'created': Timestamp('2018-01-04 01:01:01'), 'resolved': NaT,                              'priority': 'Medium', 'environment': 'PROD', 'type': None},
        {'key': 'D-5', 'created': Timestamp('2018-02-05 01:01:01'), 'resolved': Timestamp('2018-02-20 02:02:02'), 'priority': 'High',   'environment': 'SIT',  'type': None},
        {'key': 'D-6', 'created': Timestamp('2018-03-06 01:01:01'), 'resolved': NaT,                              'priority': 'Medium', 'environment': 'UAT',  'type': None},
    ]

def test_no_environment_field(jira, settings):
    settings = extend_dict(settings, {
        'defects_environment_field': None
    })

    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = DefectsCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'D-1', 'created': Timestamp('2018-01-01 01:01:01'), 'resolved': Timestamp('2018-03-20 02:02:02'), 'priority': 'High',   'environment': None, 'type': 'Config'},
        {'key': 'D-2', 'created': Timestamp('2018-01-02 01:01:01'), 'resolved': Timestamp('2018-01-20 02:02:02'), 'priority': 'Medium', 'environment': None, 'type': 'Config'},
        {'key': 'D-3', 'created': Timestamp('2018-02-03 01:01:01'), 'resolved': Timestamp('2018-03-20 02:02:02'), 'priority': 'High',   'environment': None, 'type': 'Config'},
        {'key': 'D-4', 'created': Timestamp('2018-01-04 01:01:01'), 'resolved': NaT,                              'priority': 'Medium', 'environment': None, 'type': 'Data'},
        {'key': 'D-5', 'created': Timestamp('2018-02-05 01:01:01'), 'resolved': Timestamp('2018-02-20 02:02:02'), 'priority': 'High',   'environment': None, 'type': 'Data'},
        {'key': 'D-6', 'created': Timestamp('2018-03-06 01:01:01'), 'resolved': NaT,                              'priority': 'Medium', 'environment': None, 'type': 'Data'},
    ]



import logging
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from ..calculator import Calculator
from ..utils import get_extension, set_chart_style

from .cycletime import CycleTimeCalculator

logger = logging.getLogger(__name__)

class ScatterplotCalculator(Calculator):
    """Build scatterplot data for the cycle times: a data frame containing
    only those items in where values are set for `completed_timestamp` and
    `cycle_time`, and with those two columns as the first two, both
    normalised to whole days, and with `completed_timestamp` renamed to
    `completed_date`.
    """

    def run(self):
        cycle_data = self.get_result(CycleTimeCalculator)
        columns = list(cycle_data.columns)
        columns.remove('cycle_time')
        columns.remove('completed_timestamp')
        columns.remove('blocked_days')
        columns.remove('impediments')
        columns = ['completed_timestamp', 'cycle_time', 'blocked_days'] + columns

        data = (
            cycle_data[columns]
            .dropna(subset=['cycle_time', 'completed_timestamp'])
            .rename(columns={'completed_timestamp': 'completed_date'})
        )

        data['cycle_time'] = data['cycle_time'].astype('timedelta64[D]')

        return data
    
    def write(self):
        data = self.get_result()

        if self.settings['scatterplot_data']:
            self.write_file(data, self.settings['scatterplot_data'])
        else:
            logger.debug("No output file specified for scatterplot data")
        
        if self.settings['scatterplot_chart']:
            self.write_chart(data, self.settings['scatterplot_chart'])
        else:
            logger.debug("No output file specified for scatterplot chart")

    def write_file(self, data, output_file):
        output_extension = get_extension(output_file)

        file_data = data.copy()
        file_data['completed_date'] = file_data['completed_date'].map(pd.Timestamp.date)

        logger.info("Writing scatterplot data to %s", output_file)
        if output_extension == '.json':
            file_data.to_json(output_file, date_format='iso')
        elif output_extension == '.xlsx':
            file_data.to_excel(output_file, 'Scatter', index=False)
        else:
            file_data.to_csv(output_file, index=False)
        
    def write_chart(self, data, output_file):
        if len(data.index) < 2:
            logger.warning("Need at least 2 completed items to draw scatterplot")
            return
            
        chart_data = pd.DataFrame({
            'completed_date': data['completed_date'].values.astype('datetime64[D]'),
            'cycle_time': data['cycle_time']
        }, index=data.index)

        window = self.settings['scatterplot_window']
        if window:
            start = chart_data['completed_date'].max().normalize() - pd.Timedelta(window, 'D')
            chart_data = chart_data[chart_data.completed_date >= start]

            if len(data.index) < 2:
                logger.warning("Need at least 2 completed items to draw scatterplot")
                return
        
        quantiles = self.settings['quantiles']
        logger.debug("Showing forecast at quantiles %s", ', '.join(['%.2f' % (q * 100.0) for q in quantiles]))
        
        fig, ax = plt.subplots()
        fig.autofmt_xdate()

        ax.set_xlabel("Completed date")
        ax.set_ylabel("Cycle time (days)")

        if self.settings['scatterplot_chart_title']:
            ax.set_title(self.settings['scatterplot_chart_title'])

        ax.plot_date(x=chart_data['completed_date'], y=chart_data['cycle_time'], ms=5)
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%d/%m/%Y'))

        # Add quantiles
        left, right = ax.get_xlim()
        for quantile, value in chart_data['cycle_time'].quantile(quantiles).iteritems():
            ax.hlines(value, left, right, linestyles='--', linewidths=1)
            ax.annotate("%.0f%% (%.0f days)" % ((quantile * 100), value,),
                xy=(left, value),
                xytext=(left, value + 0.5),
                fontsize="x-small",
                ha="left"
            )

        set_chart_style()

        # Write file
        logger.info("Writing scatterplot chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import logging
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from ..calculator import Calculator
from ..utils import get_extension, set_chart_style

from .cycletime import CycleTimeCalculator

logger = logging.getLogger(__name__)

class CFDCalculator(Calculator):
    """Create the data to build a cumulative flow diagram: a DataFrame,
    indexed by day, with columns containing cumulative counts for each
    of the items in the configured cycle.

    In addition, a column called `cycle_time` contains the approximate
    average cycle time of that day based on the first "accepted" status
    and the first "complete" status.

    Write as a data file and/or a diagram.
    """

    def run(self):
        cycle_data = self.get_result(CycleTimeCalculator)
        cycle_names = [s['name'] for s in self.settings['cycle']]

        # Build a dataframe of just the "date" columns
        cfd_data = cycle_data[cycle_names]

        # Strip out times from all dates
        cfd_data = pd.DataFrame(
            np.array(cfd_data.values, dtype='<M8[ns]').astype('<M8[D]').astype('<M8[ns]'),
            columns=cfd_data.columns,
            index=cfd_data.index
        )

        # Replace missing NaT values (happens if a status is skipped) with the subsequent timestamp
        cfd_data = cfd_data.fillna(method='bfill', axis=1)

        # Count number of times each date occurs, preserving column order
        cfd_data = pd.concat({col: cfd_data[col].value_counts() for col in cfd_data}, axis=1)[cycle_names]

        # Fill missing dates with 0 and run a cumulative sum
        cfd_data = cfd_data.fillna(0).cumsum(axis=0)

        # Reindex to make sure we have all dates
        start, end = cfd_data.index.min(), cfd_data.index.max()
        if start is not pd.NaT and end is not pd.NaT:
            cfd_data = cfd_data.reindex(pd.date_range(start, end, freq='D'), method='ffill')

        return cfd_data
    
    def write(self):
        data = self.get_result()

        if self.settings['cfd_data']:
            self.write_file(data, self.settings['cfd_data'])
        else:
            logger.debug("No output file specified for CFD file")
        
        if self.settings['cfd_chart']:
            self.write_chart(data, self.settings['cfd_chart'])
        else:
            logger.debug("No output file specified for CFD chart")

    def write_file(self, data, output_file):
        output_extension = get_extension(output_file)

        logger.info("Writing CFD data to %s", output_file)
        if output_extension == '.json':
            data.to_json(output_file, date_format='iso')
        elif output_extension == '.xlsx':
            data.to_excel(output_file, 'CFD')
        else:
            data.to_csv(output_file)
    
    def write_chart(self, data, output_file):
        if len(data.index) == 0:
            logger.warning("Cannot draw CFD with no data")
            return
        
        window = self.settings['cfd_window']
        if window:
            start = data.index.max() - pd.Timedelta(window, 'D')
            data = data[start:]
        
            # Re-check after slicing
            if len(data.index) == 0:
                logger.warning("Cannot draw CFD with no data")
                return

        fig, ax = plt.subplots()
        
        if self.settings['cfd_chart_title']:
            ax.set_title(self.settings['cfd_chart_title'])

        fig.autofmt_xdate()

        ax.set_xlabel("Date")
        ax.set_ylabel("Number of items")

        backlog_column = self.settings['backlog_column']

        if backlog_column not in data.columns:
            logger.error("Backlog column %s does not exist", backlog_column)
            return None

        data = data.drop([backlog_column], axis=1)
        data.plot.area(ax=ax, stacked=False, legend=False)

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))

        bottom = data[data.columns[-1]].min()
        top = data[data.columns[0]].max()
        ax.set_ylim(bottom=bottom, top=top)

        set_chart_style()

        # Write file
        logger.info("Writing CFD chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import pytest
import datetime

from .conftest import (
    FauxJIRA as JIRA,
    FauxIssue as Issue,
    FauxChange as Change,
    FauxFieldValue as Value
)

from .querymanager import QueryManager, IssueSnapshot
from .utils import extend_dict

@pytest.fixture
def jira(custom_fields):
    return JIRA(fields=custom_fields, issues=[
        Issue("A-1",
            summary="Issue A-1",
            issuetype=Value("Story", "story"),
            status=Value("Backlotg", "backlog"),
            resolution=None,
            created="2018-01-01 01:01:01",
            customfield_001="Team 1",
            customfield_002=Value(None, 30),
            customfield_003=Value(None, ["R2", "R3", "R4"]),
            changes=[
                Change("2018-01-02 01:01:01", [("status", "Backlog", "Next",)]),
                Change("2018-01-02 01:01:01", [("Team", "Team 2", "Team 1",)]),
                Change("2018-01-03 01:01:01", [("resolution", None, "Closed",), ("status", "Next", "Done",)]),
                Change("2018-01-04 01:01:01", [("resolution", "Closed", None,), ("status", "Done", "QA",)]),
            ],
        )
    ])

@pytest.fixture
def settings(custom_settings):
    return extend_dict(custom_settings, {})

def test_search(jira, settings):
    qm = QueryManager(jira, settings)
    assert qm.attributes_to_fields == {
        'Team': 'customfield_001',
        'Estimate': 'customfield_002',
        'Release': 'customfield_003',
    }

    issues = qm.find_issues("(filter=123)")
    assert issues == jira._issues

def test_resolve_attribute_value(jira, settings):
    qm = QueryManager(jira, settings)
    issues = qm.find_issues("(filter=123)")

    assert qm.resolve_attribute_value(issues[0], "Team") == "Team 1"
    assert qm.resolve_attribute_value(issues[0], "Estimate") == 30
    assert qm.resolve_attribute_value(issues[0], "Release") == "R3"  # due to known_value

def test_resolve_field_value(jira, settings):
    qm = QueryManager(jira, settings)
    issues = qm.find_issues("(filter=123)")

    assert qm.resolve_field_value(issues[0], "customfield_001") == "Team 1"
    assert qm.resolve_field_value(issues[0], "customfield_002") == 30
    assert qm.resolve_field_value(issues[0], "customfield_003") == "R3"  # due to known_value

def test_iter_changes(jira, settings):
    qm = QueryManager(jira, settings)
    issues = qm.find_issues("(filter=123)")
    changes = list(qm.iter_changes(issues[0], ['status', 'Team']))

    assert changes == [
        IssueSnapshot(change="status", key="A-1", date=datetime.datetime(2018, 1, 1, 1, 1, 1), from_string=None,      to_string="Backlog"),
        IssueSnapshot(change="Team",   key="A-1", date=datetime.datetime(2018, 1, 1, 1, 1, 1), from_string=None,      to_string="Team 2"),
        IssueSnapshot(change="status", key="A-1", date=datetime.datetime(2018, 1, 2, 1, 1, 1), from_string="Backlog", to_string="Next"),
        IssueSnapshot(change="Team",   key="A-1", date=datetime.datetime(2018, 1, 2, 1, 1, 1), from_string="Team 2",  to_string="Team 1"),
        IssueSnapshot(change="status", key="A-1", date=datetime.datetime(2018, 1, 3, 1, 1, 1), from_string="Next",    to_string="Done"),
        IssueSnapshot(change="status", key="A-1", date=datetime.datetime(2018, 1, 4, 1, 1, 1), from_string="Done",    to_string="QA")
    ]



import os
import argparse
import getpass
import logging

from jira import JIRA

from .config import config_to_options, CALCULATORS
from .webapp.app import app as webapp
from .querymanager import QueryManager
from .calculator import run_calculators
from .utils import set_chart_context

logger = logging.getLogger(__name__)

def configure_argument_parser():
    """Configure an ArgumentParser that manages command line options.
    """

    parser = argparse.ArgumentParser(description='Extract Agile metrics data from JIRA and produce data and charts.')

    # Basic options
    parser.add_argument('config', metavar='config.yml', nargs='?', help='Configuration file')
    parser.add_argument('-v', dest='verbose', action='store_true', help='Verbose output')
    parser.add_argument('-vv', dest='very_verbose', action='store_true', help='Even more verbose output')
    parser.add_argument('-n', metavar='N', dest='max_results', type=int, help='Only fetch N most recently updated issues')
    
    parser.add_argument('--server', metavar='127.0.0.1:8080', help='Run as a web server instead of a command line tool, on the given host and/or port. The remaining options do not apply.')

    # Output directory
    parser.add_argument('--output-directory', metavar='metrics', help="Write output files to this directory, rather than the current working directory.")

    # Connection options
    parser.add_argument('--domain', metavar='https://my.jira.com', help='JIRA domain name')
    parser.add_argument('--username', metavar='user', help='JIRA user name')
    parser.add_argument('--password', metavar='password', help='JIRA password')

    return parser

def main():
    parser = configure_argument_parser()
    args = parser.parse_args()

    if args.server:
        run_server(parser, args)
    else:
        run_command_line(parser, args)

def run_server(parser, args):
    host = None
    port = args.server
    
    if ':' in args.server:
        (host, port) = args.server.split(':')
    port = int(port)

    set_chart_context("paper")
    webapp.run(host=host, port=port)

def run_command_line(parser, args):
    if not args.config:
        parser.print_usage()
        return
    
    logging.basicConfig(
        format='%(message)s',
        level=(
            logging.DEBUG if args.very_verbose else
            logging.INFO if args.verbose else
            logging.WARNING
        )
    )

    # Configuration and settings (command line arguments override config file options)

    logger.debug("Parsing options from %s", args.config)
    with open(args.config) as config:
        options = config_to_options(config.read())

    # Allow command line arguments to override options
    override_options(options['connection'], args)
    override_options(options['settings'], args)

    # Set charting context, which determines how charts are rendered
    set_chart_context("paper")

    # Set output directory if required
    if args.output_directory:
        logger.info("Changing working directory to %s" % args.output_directory)
        os.chdir(args.output_directory)

    # Query JIRA and run calculators

    jira = get_jira_client(options['connection'])

    logger.info("Running calculators")
    query_manager = QueryManager(jira, options['settings'])
    run_calculators(CALCULATORS, query_manager, options['settings'])

def override_options(options, arguments):
    """Update `options` dict with settings from `arguments`
    with the same key.
    """
    for key in options.keys():
        if getattr(arguments, key, None) is not None:
            options[key] = getattr(arguments, key)

def get_jira_client(connection):
    url = connection['domain']
    username = connection['username']
    password = connection['password']
    jira_client_options = connection['jira_client_options']

    logger.info("Connecting to %s", url)

    if not username:
        username = input("Username: ")

    if not password:
        password = getpass.getpass("Password: ")

    options = {'server': url}
    options.update(jira_client_options)

    return JIRA(options, basic_auth=(username, password))



import pytest
import datetime
from pandas import DataFrame, DatetimeIndex, Timestamp

from .cfd import CFDCalculator
from .netflow import NetFlowChartCalculator

from ..utils import extend_dict

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
        'net_flow_frequency': 'D'
    })

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def results(query_manager, settings, large_cycle_time_results):
    
    # CFD data frame and net flow:
    #
    #              Backlog  Committed  Build  Test  Done
    # 2018-01-01     16.0        0.0    0.0   0.0   0.0     -->  0
    # 2018-01-02     17.0        9.0    0.0   0.0   0.0     --> +9
    # 2018-01-03     18.0       13.0    8.0   0.0   0.0     --> +4
    # 2018-01-04     18.0       14.0   10.0   3.0   0.0     --> +1
    # 2018-01-05     18.0       15.0   11.0   8.0   0.0     --> +1
    # 2018-01-06     18.0       15.0   12.0   8.0   0.0     -->  0
    # 2018-01-07     18.0       15.0   12.0   8.0   2.0     --> -2
    # 2018-01-08     18.0       15.0   12.0   9.0   4.0     --> -4
    # 2018-01-09     18.0       15.0   12.0   9.0   6.0     --> -6
    #

    return extend_dict(large_cycle_time_results, {
        CFDCalculator: CFDCalculator(query_manager, settings, large_cycle_time_results).run()
    })

def test_empty(query_manager, settings, minimal_cycle_time_columns):
    results = {
        CFDCalculator: DataFrame([], columns=['Backlog', 'Committed', 'Build', 'Test', 'Done'], index=DatetimeIndex(start=datetime.date(2018, 1, 1), periods=0, freq='D'))
    }

    calculator = NetFlowChartCalculator(query_manager, settings, results)

    data = calculator.run()
    assert list(data.columns) == ['Committed', 'Done', 'arrivals', 'departures', 'net_flow', 'positive']
    assert len(data.index) == 0


def test_columns(query_manager, settings, results):
    calculator = NetFlowChartCalculator(query_manager, settings, results)

    data = calculator.run()
    assert list(data.columns) == ['Committed', 'Done', 'arrivals', 'departures', 'net_flow', 'positive']

def test_calculate_net_flow(query_manager, settings, results):
    calculator = NetFlowChartCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.index) == [
        Timestamp('2018-01-01 00:00:00', freq='D'),
        Timestamp('2018-01-02 00:00:00', freq='D'),
        Timestamp('2018-01-03 00:00:00', freq='D'),
        Timestamp('2018-01-04 00:00:00', freq='D'),
        Timestamp('2018-01-05 00:00:00', freq='D'),
        Timestamp('2018-01-06 00:00:00', freq='D'),
        Timestamp('2018-01-07 00:00:00', freq='D'),
        Timestamp('2018-01-08 00:00:00', freq='D'),
        Timestamp('2018-01-09 00:00:00', freq='D')
    ]

    assert data[['arrivals', 'departures', 'net_flow', 'positive']].to_dict('records') == [
        {'arrivals': 0.0, 'departures': 0.0, 'net_flow':  0.0, 'positive': True},
        {'arrivals': 9.0, 'departures': 0.0, 'net_flow':  9.0, 'positive': True},
        {'arrivals': 4.0, 'departures': 0.0, 'net_flow':  4.0, 'positive': True},
        {'arrivals': 1.0, 'departures': 0.0, 'net_flow':  1.0, 'positive': True},
        {'arrivals': 1.0, 'departures': 0.0, 'net_flow':  1.0, 'positive': True},
        {'arrivals': 0.0, 'departures': 0.0, 'net_flow':  0.0, 'positive': True},
        {'arrivals': 0.0, 'departures': 2.0, 'net_flow': -2.0, 'positive': False},
        {'arrivals': 0.0, 'departures': 2.0, 'net_flow': -2.0, 'positive': False},
        {'arrivals': 0.0, 'departures': 2.0, 'net_flow': -2.0, 'positive': False},
    ]

def test_calculate_net_flow_different_columns(query_manager, settings, results):

    settings.update({
        'committed_column': 'Build',
        'done_column': 'Test',
    })

    calculator = NetFlowChartCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.index) == [
        Timestamp('2018-01-01 00:00:00', freq='D'),
        Timestamp('2018-01-02 00:00:00', freq='D'),
        Timestamp('2018-01-03 00:00:00', freq='D'),
        Timestamp('2018-01-04 00:00:00', freq='D'),
        Timestamp('2018-01-05 00:00:00', freq='D'),
        Timestamp('2018-01-06 00:00:00', freq='D'),
        Timestamp('2018-01-07 00:00:00', freq='D'),
        Timestamp('2018-01-08 00:00:00', freq='D'),
        Timestamp('2018-01-09 00:00:00', freq='D')
    ]

    assert data[['arrivals', 'departures', 'net_flow', 'positive']].to_dict('records') == [
        {'arrivals': 0.0, 'departures': 0.0, 'net_flow':  0.0, 'positive': True},
        {'arrivals': 0.0, 'departures': 0.0, 'net_flow':  0.0, 'positive': True},
        {'arrivals': 8.0, 'departures': 0.0, 'net_flow':  8.0, 'positive': True},
        {'arrivals': 2.0, 'departures': 3.0, 'net_flow': -1.0, 'positive': False},
        {'arrivals': 1.0, 'departures': 5.0, 'net_flow': -4.0, 'positive': False},
        {'arrivals': 1.0, 'departures': 0.0, 'net_flow':  1.0, 'positive': True},
        {'arrivals': 0.0, 'departures': 0.0, 'net_flow':  0.0, 'positive': True},
        {'arrivals': 0.0, 'departures': 1.0, 'net_flow': -1.0, 'positive': False},
        {'arrivals': 0.0, 'departures': 0.0, 'net_flow':  0.0, 'positive': True},
    ]



import pytest
from pandas import DataFrame, Timestamp

from .cycletime import CycleTimeCalculator
from .scatterplot import ScatterplotCalculator

from ..utils import extend_dict

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {})

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def results(large_cycle_time_results):
    return extend_dict(large_cycle_time_results, {})

def test_empty(query_manager, settings, minimal_cycle_time_columns):
    results = {
        CycleTimeCalculator: DataFrame([], columns=minimal_cycle_time_columns, index=[])
    }

    calculator = ScatterplotCalculator(query_manager, settings, results)

    data = calculator.run()
    assert list(data.columns) == [
        'completed_date',
        'cycle_time',
        'blocked_days',
        'key',
        'url',
        'issue_type',
        'summary',
        'status',
        'resolution',
        'Backlog',
        'Committed',
        'Build',
        'Test',
        'Done'
    ]
    assert len(data.index) == 0


def test_columns(query_manager, settings, results):
    calculator = ScatterplotCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.columns) == [
        'completed_date',
        'cycle_time',
        'blocked_days',
        'key',
        'url',
        'issue_type',
        'summary',
        'status',
        'resolution',
        'Backlog',
        'Committed',
        'Build',
        'Test',
        'Done'
    ]

def test_calculate_scatterplot(query_manager, settings, results):
    calculator = ScatterplotCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data[['key', 'completed_date', 'cycle_time']].to_dict('records') == [
        {'key': 'A-13', 'completed_date': Timestamp('2018-01-07 01:01:01'), 'cycle_time': 5.0},
        {'key': 'A-14', 'completed_date': Timestamp('2018-01-07 01:01:01'), 'cycle_time': 5.0},
        {'key': 'A-15', 'completed_date': Timestamp('2018-01-08 01:01:01'), 'cycle_time': 5.0},
        {'key': 'A-16', 'completed_date': Timestamp('2018-01-08 01:01:01'), 'cycle_time': 5.0},
        {'key': 'A-17', 'completed_date': Timestamp('2018-01-09 01:01:01'), 'cycle_time': 5.0},
        {'key': 'A-18', 'completed_date': Timestamp('2018-01-09 01:01:01'), 'cycle_time': 4.0},
    ]



import pytest
import datetime
from pandas import DataFrame

from .cycletime import CycleTimeCalculator
from .ageingwip import AgeingWIPChartCalculator

from ..utils import extend_dict

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
        'ageing_wip_chart': 'ageingwip.png'  # without a file to write the calculator will stop
    })

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def results(large_cycle_time_results):
    return extend_dict(large_cycle_time_results, {})

@pytest.fixture
def today():
    return datetime.date(2018, 1, 10)

def test_empty(query_manager, settings, minimal_cycle_time_columns, today):
    results = {
        CycleTimeCalculator: DataFrame([], columns=minimal_cycle_time_columns, index=[])
    }

    calculator = AgeingWIPChartCalculator(query_manager, settings, results)

    data = calculator.run(today)
    assert list(data.columns) == [
        'key',
        'summary',
        'status',
        'age',
        'Committed',
        'Build',
        'Test'
    ]
    assert len(data.index) == 0


def test_columns(query_manager, settings, results, today):
    calculator = AgeingWIPChartCalculator(query_manager, settings, results)

    data = calculator.run(today)

    assert list(data.columns) == [
        'key',
        'summary',
        'status',
        'age',
        'Committed',
        'Build',
        'Test'
    ]

def test_calculate_ageing_wip(query_manager, settings, results, today):
    calculator = AgeingWIPChartCalculator(query_manager, settings, results)

    data = calculator.run(today)

    assert data[['key', 'status', 'age']].to_dict('records') == [
        {'key': 'A-4', 'status': 'Committed', 'age': 8.0},
        {'key': 'A-5', 'status': 'Committed', 'age': 7.0},
        {'key': 'A-6', 'status': 'Committed', 'age': 7.0},
        {'key': 'A-7', 'status': 'Build', 'age': 8.0},
        {'key': 'A-8', 'status': 'Build', 'age': 8.0},
        {'key': 'A-9', 'status': 'Build', 'age': 8.0},
        {'key': 'A-10', 'status': 'Test', 'age': 8.0},
        {'key': 'A-11', 'status': 'Test', 'age': 8.0},
        {'key': 'A-12', 'status': 'Test', 'age': 8.0},
    ]

def test_calculate_ageing_wip_with_different_columns(query_manager, settings, results, today):
    settings.update({
        'committed_column': 'Committed',
        'final_column': 'Build',
        'done_column': 'Test',
    })

    calculator = AgeingWIPChartCalculator(query_manager, settings, results)

    data = calculator.run(today)

    assert data[['key', 'status', 'age']].to_dict('records') == [
        {'key': 'A-4', 'status': 'Committed', 'age': 8.0},
        {'key': 'A-5', 'status': 'Committed', 'age': 7.0},
        {'key': 'A-6', 'status': 'Committed', 'age': 7.0},
        {'key': 'A-7', 'status': 'Build', 'age': 8.0},
        {'key': 'A-8', 'status': 'Build', 'age': 8.0},
        {'key': 'A-9', 'status': 'Build', 'age': 8.0}
    ]
